/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor;

import java.io.File;
import java.io.IOException;
import java.net.*;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.extractor.metadata.MetaDataManager;
import org.xquark.extractor.oracle.OracleQueryFactory;
import org.xquark.extractor.runtime.ExtractorStatement;
import org.xquark.extractor.runtime.PreparedExtractorStatement;
import org.xquark.extractor.runtime.QueryFactory;
import org.xquark.schema.Schema;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.metadata.resolver.MetadataAccess;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryModule;
import org.xquark.xquery.xdbc.XMLDataSourceMetaDataConnectionWrapper;

/**
 * Extractor implementation of {@link org.xquark.xml.xdbc.XMLConnection}.
 */
public class ExtractorConnection extends DefaultReadOnlyXMLConnection implements XMLConnection {

    private static final String RCSRevision = "$Revision: 1.20 $";

    private static final String RCSName = "$Name:  $";

    private Connection jdbcConnection;

    private Extractor extractor = null;

    private MetaDataManager metaDataManager = null;

    private XMLDataSourceMetaData metaData = null;

    /* */
    private boolean isClosed = false;

    private String baseURI = null;

    /* a list all active XMLStatement that this XMLConnection has created */
    private List activeStatementList = new ArrayList();

    private int stmtCounter = 0; // TODO: recycle

    ExtractorConnection(Extractor extractor, Connection jdbcConnection) {
        this.extractor = extractor;
        this.jdbcConnection = jdbcConnection;
    }

    public Connection getJdbcConnection() {
        return jdbcConnection;
    }

    public Extractor getExtractor() {
        return extractor;
    }

    public synchronized void close() throws XMLDBCException {
        if (!isClosed) {
            try {
                isClosed = true;
                XMLStatement statement;
                for (int i = activeStatementList.size() - 1; i > -1; i--) {
                    statement = (XMLStatement) activeStatementList.get(i);
                    if (statement != null)
                        statement.close();
                }
                if (!jdbcConnection.isClosed())
                    jdbcConnection.close();
            } catch (SQLException ex) {
                throw new XMLDBCException("Error while closing connection", ex);
            }
        }
    }

    public void commit() throws XMLDBCException, XMLDBCNotSupportedException {
        if (!isClosed) {
            try {
                jdbcConnection.commit();
            } catch (SQLException ex) {
                throw new XMLDBCException(ex.getMessage(), ex);
            }
        } else
            throw new XMLDBCException("XMLConnection is already closed.");
    }

    public boolean isClosed() throws XMLDBCException {
        return isClosed;
    }

    public String getUserName() throws XMLDBCException {
        return extractor.getUserName();
    }

    public String getURL() throws XMLDBCException {
        return extractor.getURL();
    }

    private QueryFactory createQueryFactory(String baseURI) throws XMLDBCException {
        if (isClosed())
            throw new XMLDBCException(MessageLibrary.getMessage("CON_CLOSED"));

        QueryFactory queryFactory = null;
        String dbName = extractor.getDatabaseName();
        String dbVersion = extractor.getDatabaseVersion();

        /* DBMS version release switch (Statement implementations */
        if (-1 < dbName.indexOf("Oracle")) {
            if (-1 < dbVersion.indexOf("Release 8.1"))
                queryFactory = new OracleQueryFactory(getMetaDataManager(), getMetadataAccess());
            else if (-1 < dbVersion.indexOf("Release 7.3.4"))
                queryFactory = new org.xquark.extractor.oracle.Oracle7QueryFactory(getMetaDataManager(), getMetadataAccess());
            else
                // default Oracle is Oracle 8i
                queryFactory = new org.xquark.extractor.oracle.OracleQueryFactory(getMetaDataManager(), getMetadataAccess());
        } else if (-1 < dbName.indexOf("Microsoft"))
            queryFactory = new org.xquark.extractor.microsoft.MicrosoftQueryFactory(getMetaDataManager(), getMetadataAccess());
        else if (-1 < dbName.indexOf("Adaptive Server Enterprise"))
            queryFactory = new org.xquark.extractor.sybase.SybaseQueryFactory(getMetaDataManager(), getMetadataAccess());
        else if (-1 < dbName.indexOf("PROGRESS"))
            queryFactory = new org.xquark.extractor.progress.ProgressQueryFactory(getMetaDataManager(), getMetadataAccess());
        else if (-1 < dbName.indexOf("MySQL"))
            queryFactory = new org.xquark.extractor.mysql.MysqlQueryFactory(getMetaDataManager(), getMetadataAccess());
        else
            throw new XMLDBCNotSupportedException(MessageLibrary.getMessage("D_N_SUP_DB_VENDOR", dbName));

        queryFactory.setBaseURI(baseURI);
        return queryFactory;
    }

    public XMLStatement createStatement(short queryType) throws XMLDBCException {
        XMLStatement retVal = new ExtractorStatement(this, stmtCounter++, createQueryFactory(baseURI), getMetaDataManager());
        synchronized (this) {
            activeStatementList.add(retVal);
        }
        return retVal;
    }

    public XMLStatement createStatement() throws XMLDBCException {
        return createStatement(XQUERY_STRING_TYPE);
    }

    public PreparedXMLStatement prepareStatement(String query) throws XMLDBCException, XMLDBCNotSupportedException {
        PreparedXMLStatement retVal = new PreparedExtractorStatement(this, stmtCounter++, createQueryFactory(baseURI), getMetaDataManager(), query, extractor.getCachedStatements());
        synchronized (this) {
            activeStatementList.add(retVal);
        }
        return retVal;
    }

    /**
     * Gets the metadata regarding this connection's data source. A connection's data source is able to provide information describing itself.
     * 
     * @return an XMLDataSourceMetaData object for this connection.
     * @throws XMLDBCException
     *             if a data source access error occurs.
     */
    public XMLDataSourceMetaData getMetaData() throws XMLDBCException {
        return getMetaData(false);
    }

    public XMLDataSourceMetaData getMetaData(boolean refresh) throws XMLDBCException {
        if (metaData == null || refresh) {
            metaData = new XMLDataSourceMetaDataConnectionWrapper(this, extractor.getDataSourceMetaData(jdbcConnection, refresh, baseURI));
            metaDataManager = extractor.getMetaDataManager(jdbcConnection, false, baseURI);
        }
        return metaData;
    }

    MetaDataManager getMetaDataManager() throws XMLDBCException {
        return getMetaDataManager(false);
    }

    MetaDataManager getMetaDataManager(boolean refresh) throws XMLDBCException {
        if (metaDataManager == null || refresh) {
            metaDataManager = extractor.getMetaDataManager(jdbcConnection, refresh, baseURI);
            metaData = new XMLDataSourceMetaDataConnectionWrapper(this, extractor.getDataSourceMetaData(jdbcConnection, false, baseURI));
        }
        return metaDataManager;
    }

    MetadataAccess getMetadataAccess() throws XMLDBCException {
        return extractor.getDataSourceMetaData(jdbcConnection, false, baseURI);
    }

    public synchronized void statementClosed(XMLStatement closedStatement) throws XMLDBCException {
        activeStatementList.remove(closedStatement);
    }

    protected void finalize() throws Exception {
        close();
    }

    public XQueryModule loadModule(InputSource source) throws XMLDBCException {
        try {
            return getMetadataAccess().getModuleManager().loadModule(source, getMetadataAccess(), getMetadataAccess().getSchemaManager());
        } catch (XQueryException xqe) {
            throw new XMLDBCException("Could not read module file " + source.getSystemId() + " : " + xqe.getMessage(), xqe);
        }
    }

    public XQueryModule loadModule(String moduleURL) throws XMLDBCException {
        try {
            URL url = new URL(moduleURL);
            InputSource source = new InputSource(url.openStream());
            source.setSystemId(url.toString());
            return loadModule(source);
        } catch (IOException xqe) {
            throw new XMLDBCException("Could not read module file " + moduleURL + " : " + xqe.getMessage(), xqe);
        }
    }

    public Schema loadSchema(InputSource source) throws XMLDBCException {
        try {
            return getMetadataAccess().getSchemaManager().loadSchema(source);
        } catch (SAXException saxe) {
            throw new XMLDBCException("Could not read schema file " + source.getSystemId() + " : " + saxe.getMessage(), saxe);
        } catch (XMLDBCException xdbce) {
            throw new XMLDBCException("Could not read schema file " + source.getSystemId() + " : " + xdbce.getMessage(), xdbce);
        }
    }

    /**
     * Sets the base URI for documents.
     * 
     * @return nothing
     */
    public void setBaseURI(String baseURI) throws XMLDBCException {
        try {
            if (baseURI != null)
                if (new URI(baseURI).isAbsolute() == false)
                    throw new XMLDBCException("Incorrect base URI : " + baseURI + ", should be absolute.");
        } catch (URISyntaxException e) {
            throw new XMLDBCException("Incorrect base URI : " + baseURI, e);
        }
        this.baseURI = baseURI;
    }

}
