/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * LogicalStructureBuilder.java
 *
 * Created on 25 juillet 2001, 11:45
 */

package org.xquark.mapper.storage;

import java.io.ByteArrayOutputStream;
import java.io.UnsupportedEncodingException;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.*;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.dbms.JDBCBatcher;
import org.xquark.mapper.metadata.Node;
import org.xquark.mapper.metadata.RepositoryConstants;
import org.xquark.mapper.metadata.StoragePathMetadata;
import org.xquark.mapper.util.RecyclingStack;
import org.xquark.schema.SchemaConstants;
import org.xquark.schema.validation.ElementPSVInfoset;
import org.xquark.schema.validation.PSVInfoset;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.serialize.XMLSerializer;
import org.xquark.util.*;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLErrorHandler;
import org.xquark.xpath.NodeKind;

/**
 *
 */
public abstract class AbstractModelBuilder extends DefaultHandler
implements RepositoryConstants, _StorageContext, RecyclingStack.StackObjectFactory
{
    private static final String RCSRevision = "$Revision: 1.8 $";
    private static final String RCSName     = "$Name:  $";
    
    /** Backing stack
     */
    protected RecyclingStack stack = null;
    
    protected StorageBuffer storageBuffer;  // Class buffering data storage to the database.
    protected AbstractConnection connection;
    protected JDBCBatcher batcher;
    protected SchemaValidationContext validationContext;
    /** Handlers used for building XML fragments, stored in CLOBs */
    protected ArrayList contentHandlers = null;
    
    private boolean autoFlush = true;
    
    public AbstractModelBuilder(AbstractConnection connection, SchemaValidationContext schemaContext)
    throws XMLDBCException
    {
        stack = new RecyclingStack(this);
        this.connection = connection;
        validationContext = schemaContext;
        batcher = new JDBCBatcher(connection, autoFlush);
        contentHandlers = new ArrayList();
     }
    
    protected abstract AbstractModelNode getLogicalStructureNode(byte type, StoragePathMetadata path, String data, Locator locator)
    throws SQLException, XMLDBCException;
    
    protected abstract AbstractModelNode finalizeNode(Locator locator)
    throws XMLDBCException, SQLException;
    
    /* return the last OID allocated */
    public void endModel() throws XMLDBCException, SQLException
    {
        // Flush completed tuples
        storageBuffer.flush();
    }
    
    public void reset()
    throws RepositoryException, SQLException
    {
        stack.clear();
        storageBuffer.reset();
        contentHandlers.clear();
    }
    
    public void close() throws RepositoryException, SQLException
    {
        if (batcher != null)
        {
            batcher.close();
            batcher = null;
            storageBuffer.close();
            storageBuffer = null;
            stack = null;
        }
    }
    
    public void setErrorHandler(XMLErrorHandler handler)
    {
        storageBuffer.setErrorHandler(handler);
    }
    
    public XMLErrorHandler getXMLErrorHandler()
    {
        return storageBuffer.getXMLErrorHandler();
    }
    
    public AbstractModelNode getCurrentNode()
    {
        return (AbstractModelNode)stack.top();
    }
    
    ////////////////////////////////////////////////////////////////////
    //              Common features.
    ////////////////////////////////////////////////////////////////////
    /**
     * For root or element nodes.
     **/
    public AbstractModelNode newNode(StoragePathMetadata path, Locator locator)
    throws SQLException, XMLDBCException
    {
        if (path == null)
            return null;
        return getLogicalStructureNode(path.getType(), path, "", locator);  // "" ant not null because empty element will never have setData() called...
    }
    
    /**
     * For Text nodes, attributes
     **/
    public AbstractModelNode leafNode(byte type, StoragePathMetadata path, String data, Locator locator)
    throws XMLDBCException, SQLException
    {
    	getLogicalStructureNode(type, path, data, locator);
    	return finalizeNode(locator);
    }
    
    public void setData(String data)
    {
        getCurrentNode().setData(data);
    }
    
    public void setXSIinfo(String attName, String value)
    throws RepositoryException
    {
        getCurrentNode().setXSIinfo(attName, value);
    }
    
    public void incCharOffset(int length)
    {
        getCurrentNode().incCharOffset(length);
    }
    
    public boolean empty()
    {
        return stack.isEmpty();
    }
    
    public void flushBuffer() throws XMLDBCException
    {
        batcher.flush();
    }
    
    public void clearBuffer() throws XMLDBCException
    {
        try {
            batcher.reset();
        }
        catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR, 
            "Could not reset the JDBC Batcher");
        }
    }
    
    public void setAutoFlush(boolean mode) throws XMLDBCException
    {
        autoFlush = mode;
        batcher.setAutoFlush(mode);
    }
    
    public boolean getAutoFlush()
    {
        return autoFlush;
    }
    ////////////////////////////////////////////////////////////////////
    //          IMPLEMENTATION of StorageContext interface.
    ////////////////////////////////////////////////////////////////////
    public long getBucketOID() { return -1;}
    public short getPathOID() { return -1;}
    public String getDocumentID() { return null;}
    public long getDocumentOID() { return -1;}
    public long getOID() { return -1;}
    public long getUOID() { return -1;}
    
    public String getQName()
    {
        AbstractModelNode node = getCurrentNode();
        return (node.getPath().getNamespace() == null ? node.getPath().getLocalName() : node.getPath().getNamespace() + ":" + node.getPath().getLocalName());
    }
    public int getNodeRank()
    {
        return getCurrentNode().rank;
    }
    public String getNodeData() { return getCurrentNode().data;}
    public String getNormalizedNodeData()
    {
        String ret = null;
        PSVInfoset nodeInfoSet = getNodeInfoSet();
        if (nodeInfoSet == null)
            return null;

        // do not take default value coming from schema if node is an attribute
        if ((nodeInfoSet.getSpecified() == null) || // no declaration
            (
            (nodeInfoSet.getSpecified() == PSVInfoset.SCHEMA) // constant equality
                && (getCurrentNode().getPath().getType() == NodeKind.ATTRIBUTE)
            ))
            ret = getNodeData();
        else
            ret = nodeInfoSet.getNormalizedValue();
        
        return ret;
    }
    
    public Object getActualNodeData()
    {
        Object ret = null;
        PSVInfoset nodeInfoSet = getNodeInfoSet();
        if (nodeInfoSet == null)
            return null;

         // do not take default value coming from schema if node is an attribute
        if ((nodeInfoSet.getSpecified() == null) || // no declaration
            (
            (nodeInfoSet.getSpecified() == PSVInfoset.SCHEMA) // constant equality
                && (getCurrentNode().getPath().getType() == NodeKind.ATTRIBUTE)
            ))
            ret = getNodeData();
        else
            ret = nodeInfoSet.getActualValue();  // default value
        
        return ret;
   }
    public String getCanonicalString()
    {
        String ret = null;
        PSVInfoset nodeInfoSet = getNodeInfoSet();
        if (nodeInfoSet == null)
            return null;

        // do not take default value coming from schema if node is an attribute
        if ((nodeInfoSet.getSpecified() == null) || // no declaration
        (
        (nodeInfoSet.getSpecified() == PSVInfoset.SCHEMA) // constant equality
        && (getCurrentNode().getPath().getType() == NodeKind.ATTRIBUTE)
        ))
            ret = getNodeData();
        else
            ret = nodeInfoSet.getDeclaration().getType().getValueType().toXMLString(
                nodeInfoSet.getActualValue(), null);
        
        return ret;
    }
    public String getLocalName() { return getCurrentNode().getPath().getLocalName();}
    public String getNamespaceURI() { return getCurrentNode().getPath().getNamespace();}
    public Object getRefValue(int tableIndex, int columnIndex)
    {
    	StorageBuffer.Tuple master = storageBuffer.getWorkingTuple().getMasterTuple(tableIndex);
    	if (master != null) return master.getParameter(columnIndex);
    	else return null;
    }
    public Connection getConnection()
    {
        return storageBuffer.getConnection().getConnection();
    }
    
    public String getMappingNamespace() {
    	PSVInfoset infoset = validationContext.getInfoset(1);
    	return infoset.getNamespaceURI();
    }
    
    public String getMappingMode() {
    	PSVInfoset infoset = validationContext.getInfoset(1);
    	return infoset.getLocalName();
    }
    
    public RecyclingStack.StackObject newStackObject()
    {
        return new AbstractModelNode();
    }
    
    //
    //                  PRIVATE UTILITIES
    //
    
    private PSVInfoset getNodeInfoSet()
    {
        AbstractModelBuilder.AbstractModelNode wNode = getCurrentNode();
        if (wNode.data == null)
            return null;
        
        StoragePathMetadata node = wNode.getPath();
        ElementPSVInfoset infoSet = validationContext.getCurrentInfoset();
        
        PSVInfoset nodeInfoSet = null;
        if (node.getType() == NodeKind.ELEMENT)
            nodeInfoSet = infoSet;
        else    // attribute
            nodeInfoSet = infoSet.getAttributePSVInfoset(node.getNamespace(), node.getLocalName());
        
        return nodeInfoSet;
    }
    
    //
    //  XML fragments construction
    //
    
    public void startFragmentRecording() {
    	XMLFragmentRecorder recorder = new XMLFragmentRecorder(validationContext);
    	contentHandlers.add(recorder);
    }
    
    public String getRecordedFragment() {
    	XMLFragmentRecorder recorder = (XMLFragmentRecorder)contentHandlers.remove(contentHandlers.size()-1);
    	return recorder.getFragmentAsString();
    }
    
	public void startElement(String namespaceURI, String localName, String qName, Attributes attributes)
	throws SAXException
	{
		for (int i = 0; i < contentHandlers.size(); i++) 
			((ContentHandler)contentHandlers.get(i)).startElement(namespaceURI, localName, qName, attributes);
	}
    
	public void characters(char[] data, int start, int length)
	throws SAXException
	{
		for (int i = 0; i < contentHandlers.size(); i++) 
			((ContentHandler)contentHandlers.get(i)).characters(data, start, length);
	}
    
	public void endElement(String namespaceURI, String localName, String qName)
	throws SAXException
	{
		for (int i = 0; i < contentHandlers.size(); i++) 
			((ContentHandler)contentHandlers.get(i)).endElement(namespaceURI, localName, qName);
	}
    
	public void startPrefixMapping(String prefix, String uri) throws SAXException
	{
		for (int i = 0; i < contentHandlers.size(); i++) 
			((ContentHandler)contentHandlers.get(i)).startPrefixMapping(prefix, uri);
	}
    
	public void ignorableWhitespace(char[] data, int start, int length)
	throws SAXException
	{
		for (int i = 0; i < contentHandlers.size(); i++) 
			((ContentHandler)contentHandlers.get(i)).ignorableWhitespace(data, start, length);
	}
    
	public void endPrefixMapping(String prefix)  throws SAXException
	{
		for (int i = 0; i < contentHandlers.size(); i++) 
			((ContentHandler)contentHandlers.get(i)).endPrefixMapping(prefix);
	}
	
	class XMLFragmentRecorder extends NamespaceContextHandler {
		
		ByteArrayOutputStream outStream = new ByteArrayOutputStream();
		NamespaceContext externalContext = null;
		List fragmentPrefixes = null;
		
		public XMLFragmentRecorder(NamespaceContext nsContext) {
			try {
				XMLSerializer serializer = new XMLSerializer(outStream, "UTF-8");
				serializer.setIndent(false);
				serializer.setUseIgnorableWhitespaces(false);
				serializer.setGenerateXMLDeclaration(false);
				setContentHandler(serializer);
			} catch (UnsupportedEncodingException e) {
				// should not occur
			}
			externalContext = nsContext;
			fragmentPrefixes = externalContext.getDeclaredPrefixes();
			startRecording();
		}
		
		public String getFragmentAsString() {
			String result = null;
			stopRecording();
			try {
				result = outStream.toString("UTF-8"); 
			} catch (UnsupportedEncodingException e) {
				// should not occur
			}
			return result;
		}
		
		public byte[] getFragmentAsByteArray() {
			stopRecording();
			return outStream.toByteArray();
		}

		public void startElement(String namespaceURI, String localName, String qName, Attributes attributes)
		throws SAXException
		{
			checkPrefix(namespaceURI);
			for (int i = 0; i < attributes.getLength(); i++) {
				checkPrefix(attributes.getURI(i));
			}
			super.startElement(namespaceURI, localName, qName, attributes);
		}
	    
		public void endElement(String namespaceURI, String localName, String qName)
		throws SAXException
		{
			super.endElement(namespaceURI, localName, qName);
			List localPrefixes = super.getDeclaredPrefixes();
			List externalPrefixes = externalContext.getDeclaredPrefixes();
			Iterator it = localPrefixes.iterator();
			while (it.hasNext()) {
				String prefix = (String)it.next();
				if (!externalPrefixes.contains(prefix))
					super.endPrefixMapping(prefix);
			}
		}
	    
		private void checkPrefix(String namespaceURI) throws SAXException {
			if (namespaceURI == null || namespaceURI.length() == 0)
				return;
			String prefix = super.getPrefix(namespaceURI);
			if (prefix == null) {
				prefix = externalContext.getPrefix(namespaceURI);
                if (prefix == null)
                    throw new SAXException("Prefix not found for namespace " + namespaceURI);
				super.startPrefixMapping(prefix, namespaceURI);
			}
		}
		
		private void startRecording() {
			try {
				super.startDocument();
				Iterator it = fragmentPrefixes.iterator();
				while (it.hasNext()) {
					String prefix = (String)it.next();
					super.startPrefixMapping(prefix, externalContext.getNamespaceURI(prefix));
				}
			} catch (SAXException e) {
				// should not occur
			}
		}

		private void stopRecording() {
			try {
				Iterator it = fragmentPrefixes.iterator();
				while (it.hasNext()) {
					String prefix = (String)it.next();
					super.endPrefixMapping(prefix);
				}
				super.endDocument();
			} catch (SAXException e) {
				// should not occur
			}
			setContentHandler(null);
		}
	}
    
    ////////////////////////////////////////////////////////////////////
    //                  STORAGE MODEL NODE
    ////////////////////////////////////////////////////////////////////
    public class AbstractModelNode extends Node
    implements RecyclingStack.StackObject
    {
        protected StoragePathMetadata pathMetadata;
        protected short childCount = 0; // child element count (attributes are discarded)
        private short rank = 0; // position among siblings child element or text (attributes are discarded)
        private String data = null;
        protected StorageBuffer.BufferNode tupleNode = null;
        
        AbstractModelNode()
        {}
        void set(StoragePathMetadata path, byte type, String data, short rank)
        {
            pathMetadata = path;
            this.type = type;
            this.rank = rank;
            this.path = path.getPathID();
            setData(data);
        }
        public void clear()
        {
            super.clear();
            pathMetadata = null;
            childCount = 0;
            rank = 0;
            data = null;
            tupleNode = null;
        }

        void setTupleNode(StorageBuffer.BufferNode node)
        {
            tupleNode = node;
        }
        
        StorageBuffer.BufferNode getTupleNode()
        {
            return tupleNode;
        }
        
        String getData()
        {
            return data;
        }
        
        short getRank()
        {
            return rank;
        }
        
        short incChildCount()
        {
            return ++childCount;
        }
        
        /** not leaf to the storage model i.e.,  text OR attribute OR leaf element
         */
        boolean isLeaf()
        {
            return (childCount == 0);
        }

        public void setData(String data)
        {
            this.data = data;
        }

        StoragePathMetadata getPath()
        {
            return pathMetadata;
        }
        public void incCharOffset(int length)
        {
            // no op
        }
        
        /**
         * @return true if nil.
         */
        public boolean setXSIinfo(String attName, String value)
        throws RepositoryException
        {
            if (attName.equals(SchemaConstants.XSI_NIL_ATTR)
                && (value.equals(SchemaConstants.TRUE_VALUE) || value.equals("1")))
            {
                setData(null);    // replaces "" to avoid setting "" on a Number column for instance
                return true;
            }
            return false;
        }
    }
}
